@@ -1,4 +1,5 @@
 from __future__ import annotations
+import asyncio
 from pydantic import parse_obj_as
 import logging
 import json
@@ -9,9 +10,13 @@ from gpt_router.models import (
     GPTRouterMetadata,
     ImageGenerationRequest,
     ImageGenerationResponse,
+    ImageEditRequest,
     ModelGenerationRequest,
     GenerationResponse,
     ChunkedGenerationResponse,
+    ModelRouterEmbeddingsGenerationRequest,
+    ModelRouterGenerationResponse,
+    ModelRouterEmbeddingsGenerationResponse,
 )
 from gpt_router.exceptions import (
     GPTRouterApiTimeoutError,
@@ -63,6 +68,8 @@ class GPTRouterClient:
         additional_metadata: Optional[GPTRouterMetadata] = None,
     ):
         self.base_url = base_url
+        if api_key is None:
+            raise ValueError("API key cannot be None. Please provide a valid API key.")
         self.api_key = api_key
         self.request_timeout = request_timeout
         self.additional_metadata = additional_metadata
@@ -70,6 +77,29 @@ class GPTRouterClient:
             request_timeout if stream_read_timeout is None else stream_read_timeout
         )
 
+    async def agenerate_embeddings(
+        self,
+        *,
+        ordered_generation_requests: List[ModelRouterEmbeddingsGenerationRequest],
+        model_router_metadata: Optional[GPTRouterMetadata] = None,
+    ) -> ModelRouterGenerationResponse:
+        api_payload = {
+            "data": [
+                {
+                    **request.dict(),
+                    "order": index + 1,
+                }
+                for index, request in enumerate(ordered_generation_requests)
+            ],
+        }
+        api_payload = self.add_metadata_info(api_payload, model_router_metadata)
+        response = await self._async_api_call(
+            path="/v1/generate/embeddings",
+            method="POST",
+            payload=api_payload,
+        )
+        return ModelRouterEmbeddingsGenerationResponse.parse_obj(response)
+
     def add_metadata_info(
         self,
         payload: dict,
@@ -89,6 +119,7 @@ class GPTRouterClient:
                 "historyId": (
                     str(metadata["history_id"]) if metadata.get("history_id") else None
                 ),
+                "appId": metadata.get("appId"),
             }
         )
 
@@ -153,6 +184,10 @@ class GPTRouterClient:
 
     async def astream_events(self, *, path: str, method: str, payload: dict, **kwargs):
         retry_count = kwargs.get("stream_retry_count") or 1
+        stop_stream_signal = kwargs.get("stop_stream_signal")
+        switch_models_order_on_timeout = (
+            kwargs.get("switch_models_order_on_timeout") or False
+        )
         async_retrying = AsyncRetrying(
             stop=stop_after_attempt(retry_count),
             wait=wait_fixed(0.5),
@@ -191,6 +226,10 @@ class GPTRouterClient:
                             raise ServerError("Server error")
                         response.raise_for_status()
                         async for line in response.aiter_lines():
+                            if stop_stream_signal and stop_stream_signal.is_set():
+                                logger.info("Stopping stream due to signal")
+                                await response.aclose()
+                                return
                             try:
                                 if line.strip() == "":
                                     continue
@@ -217,17 +256,28 @@ class GPTRouterClient:
                 if attempt_number == retry_count:
                     logger.error("All retry attempts failed")
                     raise
+                elif switch_models_order_on_timeout:
+                    if attempt_number % 2:
+                        self._reorder_models_on_timeout(payload)
+                    logger.info(f"Retrying as per attempt #{attempt_number}")
             except httpx.HTTPStatusError as err:
                 raise Exception(
                     f"HTTP Error {err.response.status_code}: {err.response.text}"
                 ) from err
             except httpx.TimeoutException as err:
                 raise TimeoutError("Request timed out") from err
+            finally:
+                if stop_stream_signal and stop_stream_signal.is_set():
+                    logger.info("Stream stopped due to signal")
+                    return
 
     def stream_events(
         self, *, path: str, method: str, payload: dict, **kwargs
     ) -> Generator[ChunkedGenerationResponse]:
         retry_count = kwargs.get("stream_retry_count") or 1
+        switch_models_order_on_timeout = (
+            kwargs.get("switch_models_order_on_timeout") or False
+        )
         retrying = Retrying(
             stop=stop_after_attempt(retry_count),
             wait=wait_fixed(0.5),
@@ -292,6 +342,10 @@ class GPTRouterClient:
                 if attempt_number == retry_count:
                     logger.error("All retry attempts failed")
                     raise
+                elif switch_models_order_on_timeout:
+                    if attempt_number % 2:
+                        self._reorder_models_on_timeout(payload)
+                    logger.info(f"Retrying as per attempt #{attempt_number}")
             except httpx.HTTPStatusError as err:
                 raise Exception(
                     f"HTTP Error {err.response.status_code}: {err.response.text}"
@@ -299,6 +353,30 @@ class GPTRouterClient:
             except httpx.TimeoutException as err:
                 raise TimeoutError("Request timed out") from err
 
+    def _reorder_models_on_timeout(self, payload: dict) -> None:
+        """Helper method to reorder models in payload when timeout occurs."""
+        if not payload:
+            payload = {}
+        data = payload.get("data", [])
+        try:
+            if (
+                len(data) >= 2
+                and isinstance(data[0], dict)
+                and isinstance(data[1], dict)
+            ):
+                data_0_order = data[0].get("order", None)
+                data_1_order = data[1].get("order", None)
+                if data_1_order and data_0_order:
+                    (
+                        payload["data"][0]["order"],
+                        payload["data"][1]["order"],
+                    ) = (
+                        payload["data"][1]["order"],
+                        payload["data"][0]["order"],
+                    )
+        except Exception as ex:
+            logger.exception(f"Error reordering models: {ex}")
+
     def generate(
         self,
         *,
@@ -338,6 +416,7 @@ class GPTRouterClient:
         ordered_generation_requests: List[ModelGenerationRequest],
         is_stream=False,
         model_router_metadata: Optional[GPTRouterMetadata] = None,
+        stop_stream_signal: Optional[asyncio.Event] = None,
         **kwargs,
     ) -> GenerationResponse:
         api_path = "/v1/generate"
@@ -355,6 +434,7 @@ class GPTRouterClient:
                 path=api_path,
                 method=api_method,
                 payload=api_payload,
+                stop_stream_signal=stop_stream_signal,
                 **kwargs,
             )
         result = await self._async_api_call(
@@ -379,9 +459,67 @@ class GPTRouterClient:
         )
         generated_images = api_response.get("response", [])
         if isinstance(generated_images, dict):
-            generated_images = generated_images.get("artifacts", [])
+            generated_images = self._extract_images_from_dict(generated_images)
 
         return [
             parse_obj_as(ImageGenerationResponse, generated_img)
             for generated_img in generated_images
         ]
+
+    def edit_image(self, *, image_edit_request: ImageEditRequest) -> List[ImageGenerationResponse]:
+        """
+        Edit an image based on the provided ImageEditRequest.
+        This is a synchronous wrapper around the async method.
+        
+        Args:
+            image_edit_request: The request containing the image and editing instructions
+            
+        Returns:
+            List of image generation responses
+        """
+        import asyncio
+        
+        async def _async_call():
+            return await self.aedit_image(image_edit_request=image_edit_request)
+            
+        return asyncio.run(_async_call())
+    
+    async def aedit_image(self, *, image_edit_request: ImageEditRequest) -> List[ImageGenerationResponse]:
+        """
+        Edit an image based on the provided ImageEditRequest.
+        
+        Args:
+            image_edit_request: The request containing the image and editing instructions
+            
+        Returns:
+            List of image generation responses
+        """
+        api_path = "/v1/generate/edit-image"
+        api_method = "POST"
+        api_payload = image_edit_request.dict()
+        
+        api_response = await self._async_api_call(
+            path=api_path,
+            method=api_method,
+            payload=api_payload,
+        )
+        
+        generated_images = api_response.get("response", {}).get("data", [])
+        
+        return [
+            parse_obj_as(ImageGenerationResponse, {
+                "url": img.get("url", None),
+                "base64": img.get("b64_json", None),
+                "finish_reason": "SUCCESS"
+            })
+            for img in generated_images
+        ]
+
+    def _extract_images_from_dict(self, image_dict: dict) -> List[dict]:
+        if artifacts := image_dict.get("artifacts"):
+            return artifacts
+
+        if image_urls := image_dict.get("imageUrls"):
+            return [{"url": url} for url in image_urls]
+
+        return []
